# Working with files and directories

??? quote "First click and follow the instructions below only if you start the course at this stage! Otherwise skip this step!"
    {%
    include-markdown "pages/bash/bash-0-setup.md"
    %}

Let's be sure to be in the `script` directory before to continue the rest of the training:
```bash
cd ~/{{extra.working_directory}}/script
```

![](../images/3-navigation/tree_exo5.png){: style="height:250px"}

## Creating directories (mkdir)

If we want to run an analysis, it will be nice to have an `analysis` dedicated folder.

!!! question "Create a new directory called `analysis` at the same level than `script` and `data`?"

??? example "Click to show the solution"
    `mkdir analysis`


## Creating files (touch, nano)

??? tip "Course reminder"
    Actually there are 3 main possibilities to create a file: 

    * using `touch` will create an empty file
    * using a text editor. There are plenty of text editors available: `ed`, `ex`, `vi`, `vim`, `emacs`, `nano`, etc. As beginner, a good choice would be `nano`... why? Because the command how to use it are directly discribe in the screen it opens when you call it.
    * redirecting a command output into a file using the special characters `>` or `>>`. But this is too early to learn this way for now.

    

!!! question "Create a new file called `protocol.txt` using `touch` at the same level than `script`, `data` and analysis"

??? example "Click to show the solution"
    `touch protocol.txt`

!!! question "Edit the new created file  `protocol.txt` using `nano`"

??? example "Click to show the solution"
    `nano protocol.txt`


You will end up in a window that represents the content of your file, with the cursor ready to listen what you want to write. At the bottom of the screen you will see description of few commands available. Indeed the mouse is useless here... everything has to pass by a command, which will be a combination of two characters where the first one is always `^` and corresponds to the **control key** on your keyboard.

![](../images/4-creation/nano.png){: style="height:250px"}

Try to display the help `^G`. Not easy isnt'it? 
The great thing about `**nano** is that you do not need a cheat sheet of commands, they are right there in front of you.

Let's now write something: `I love guacamole`  
and save the result with `^O`

What happens? Actually nano before to quit and save ask you if you are happy with the file name where it will write. Yes sure! So to go ahead just type `↵ Return`. 

!!! question "Why `nano` ask if we wish to keep the file name?"

??? example "Click to show the solution"
    To be sure you don't do any wrong manipulation and you are on the point to overwrite a file very important. Right, once the modification is done, there is no turning back.

You can now exit with `^X`.

!!! tip
    You can directly type `^X` when you are done with your edition. `nano` will ask first if you want to save the result, then ask if you are sure about the name of the file you are on the way to write in, and then exit.  


??? tip "Course reminder about naming conventions"

    There are few rules to respect to make your life smoother.

    * No space
    * No accent or special symbol! forbidden! `&` `~` `#` `”` `'` `{` `(` `[` `|` `\` `` ` `` `^` `@` `)` `]` `}` `$` `*` `%` `!` `/` `;` `,` `?`
    * Use roman letters, numbers and `_` and `-`
    * use file extension to guess what is inside `.fa` `.txt` `.bam` changing the extension of a fasta file by the `.bam` extension doesn’t somehow magically turn it into an alignment file. Files merely contain bytes; it’s up to us and our programs to interpret those bytes according. 
    * Unix systems are case sensitive. Choose a way to write and follow it! `Sequence.fa` ≠ `SEQUENCE.fa` ≠ `sequence.fa`

## Displaying a file (cat, more)

!!! question "Using different commands, try to display the content of the file `script1.sh` located in the `script`directory. Do you see the difference?"

??? example "Click to show the solution"  
    ```bash
    more script/script1.sh
    cat script/script1.sh
    less script/script1.sh
    ```


## Moving files (mv)

??? tip "Course reminder"
    The command `mv` allows to move file and directories. It follow this syntax:  
    `mv <source> <target>`  
    For both, the source (what we want to move) and the target (where we want to move it), it is possible to use either relative or absolute path. 

We want to move the `analysis` directory and the `protocol.txt` file in a correct location  

Below is the current situation on the left and the situation we want to end up on the right:

![](../images/4-creation/move1.png){: style="height:250px"} ![](../images/4-creation/move2.png){: style="height:250px"}

!!! question "Move the `analysis` folder and `protocol.txt` accordingly."

??? example "Click to show the solution"
    using the absolute path:
    ```bash
    mv /home/<userName>/{{extra.working_directory}}/script/analysis /home/<userName>/{{extra.working_directory}}/analysis
    mv /home/<userName>/{{extra.working_directory}}/script/protocol.txt /home/<userName>/{{extra.working_directory}}/analysis/ 
    ```
    using the relative path:  
    ```bash
    mv analysis ..
    mv protocol.txt ../analysis/
    ```
    You could even mix abolute and relative path...

## Copy files and directories (cp)

??? tip "Course reminder"
    The command to copy files and directories is `cp`. But for directories you will need to use the `-r` option that stands for recursive (copy also everything inside).
    Once again you need to well understand where you are in your directory structure and use either the absolute or relative path of the file/folder you want to copy and the place you want to paste it.  
    Remember that the current directory is also equivqlent to `.`

Let's imagine you want your script `script1.sh` locally to run your analysis, but you want just a copy. You want to keep the original file as it is. 

!!! question "Go into the `analysis` folder and copy the `script1.sh` file inside."


??? example "Click to show the solution"  
    ```bash
    cd ../analysis
    cp ../script/script1.sh .
    ```


!!! question "Copy the entire `script` folder"

??? example "Click to show the solution"  
    ```bash
    cp -r ../script .
    ```
    /!\ If you use `cp -r ../script/ .` you will copy everything that is inside but not the `script` folder itself! So you see the difference?      
    Check the result:
    ```bash
    pwd
    ls
    ls script
    ```



## Renaming files/folders (mv)

Sadly you realize your protocol is bad you will have to write another one. To keep a track of it you just want to rename it.
The easiest way to rename a file or a directory is to use the `mv` command.

!!! question "Change the `protocol.txt` file into `protocol_wrong.txt`."

??? example "Click to show the solution"  
    ```bash
    mv protocol.txt protocol_wrong.txt
    ```

If you use a full path, you just have to follow the same concept by renaming the `protocol_wrong.txt` into `protocol_very_wrong.txt` using absolute path:

`mv /home/<userName>/{{extra.working_directory}}/script/protocol_wrong.txt /home/<userName>/{{extra.working_directory}}/script/protocol_very_wrong.txt`

!!! question "Change the name of the `script` directory into `script_copy`"
??? example "Click to show the solution"  
    ```bash
    mv script script_copy
    ls -Fa # to check everything went well
    ```

## Removing files/folders (rm)

Let's clean the mess we have done in this `analysis folder`. The command to remove files and directories is `rm`. In the case of directory you will have to use the `-r` option to make it recursive.

!!! question "remove the `script1.sh` file."

??? example "Click to show the solution"  
    ```bash
    rm script1.sh
    ```

!!! question "Now remove the `fastq` files using the filename expansion."

??? example "Click to show the solution"  
    ```bash
    rm *.fastq
    ```

!!! question "Let's finish this exercice by removing the `script_copy` directory."

??? example "Click to show the solution"  
    ```bash
    rm -r script_copy
    ls -Fa # to check everything went well
    ```

## Linking files/folders (ln -s)

Actually in a real world what we have done so far was not very recommended, especially copy `fastq` files. 

!!! question "Do you guess why we should have avoid to copy those files?"

??? example "Click to show the solution"  
    Fastq files are generaly huge. Doing a copy double the disk space used by the file, because now you have it twice. It is recommended when possible to avoid to make copies of huge files.

Another situation is when you want to access files far away many times (e.g. input files for a local script) and you will have to deal with long path. Long paths can be annoying...

There is a solution to ease your life, giving you the simplicity to have a file locally (no need to think about the path), to avoid having a copy of the file and to avoid dealing with long path: The **link**!

We focus only on `soft link` also known as `symbolic link`. It is a special sort of file that points at a different file. You could think of it like a shortcut.
 Be carefull, if you remove the original file, the link become inoperable. 

![](../images/4-creation/soft_link.png){: style="height:350px"}

 The solft link is made via the `ln -s` command.

```bash
ln -s existing_file [soft_link_name]
```

!!! Exercice
    Create a Soft link of the `script` folder
    ??? example "Click to show the solution"  
        ```ln  -s ../script```
    Run `ls -l`. What do you see?
    ??? example "Click to show the solution"  
        There is an extra arrow showing to what file my link point to.
    Create a Soft link of the `script/script1.sh` file and run again `ls -l`
    ??? example "Click to show the solution"  
        ```bash
        ln -s ../data/data1.fa
        ls -l
        ```

To remove links you can use `rm` like for files or the dedicated `unlink` command (the only difference is that unlink cannot remove directories).

!!! question "Remove the links we set: `script`, `script1.sh`"

??? example "Click to show the solution"  
    ```bash
    rm script script1.sh
    ```


## Jokers/Wildcards (*)

Before running your analysis, you may be interested to access all the fastq files present in the `data` folder.
You can use the `cp` command for all the files one by one as we have seen previously. But if you have hundred of those `fastq` files, it would be fastidious.  
Well Bash does carry out a very powerful feature, the `filename expansion`, a process known as globbing! (to not confound with regular expressions). 

??? tip "Course reminder about globbing/jokers"

    **Here the list of globbing:**

    | wildcard characters  | meaning |
    | :-: | :-: |
    | `?` | Any single character |
    | `*` | Zero or more characters |
    | `[]` | Specify a range. Any character of the range or none of them by using `!` inside the bracket.|
    | `{term1,term2}` | Specify a list of terms separated by commas and each term must be a name or a wildcard.|
    | `{term1..term2}` | Called brace expansion, this syntax expands all the terms between term1 and term2 (Letters or Integers).|

    The most important one you have to remember is `*`.


!!! question "Using `ls`, list all the fastq files available in the `data` folder (and only the fastq files!)"

??? example "Click to show the solution"  
    ```bash
    ls ../data/data?.fastq 
    ```
    This one works but if we had a file called `dataPE.fastq` it would not appear because only a single character is expected between `data` and `.fastq` while here it would be `PE`.
    ```bash
    ls ../data/*.fastq 
    ```
    Excellent this one match all files ending in `.fastq`.



!!! question "Now copy all the fastq files using a single command thanks to the `*` filename expansion."

??? example "Click to show the solution"  
    ```bash
    cp ../data/*.fastq .
    ```
    Very convenient isn't it?




## Commands to keep in mind

| Action | Files | Folders |
| :-: | :-: | :-: |
| Inspect | ls | ls |
| View content | cat | ls |
| Navigate to |  | cd |
| Move | mv | mv |
| Copy | cp | cp -r |
| Create | nano/touch | mkdir |
| Delete | rm | rmdir, rm -r |

!!! Success "Quick recap"
    In this section we've learned:

    - How to create files and directories
    - How to move files and directories
    - How to rename files and directories
    - How to delete files and directories
    - How to link files and directories
